(*
 * SNU 4190.310 Programming Languages (Fall 2010)
 *
 * Let-polymorphism with M algorithm
 *)
open M

module M_PolyChecker : M_PolyChecker = struct
    let check exp = raise (TypeError "no checker") (* TODO: implementation *)
end

module M_SimChecker_M : M_SimTypeChecker = struct
    (* definition of types part *)

    type ty = TVar of int
            | TInt | TString | TBool
            | TPair of ty * ty
            | TLoc of ty
            | TArrow of ty * ty

    let next_tvar = ref 0
    let newt() = next_tvar := !next_tvar + 1; TVar !next_tvar

    let rec string_of_types =
        function TInt -> "int"
         | TString -> "string"
         | TBool -> "bool"
         | TVar x -> "'" ^ string_of_int x
         | TPair (t1, t2) ->
            "(" ^ string_of_types t1 ^ " * " ^ string_of_types t2 ^ ")"
         | TLoc t -> "loc " ^ string_of_types t
         | TArrow (t1, t2) ->
            "(" ^ string_of_types t1 ^ "->" ^ string_of_types t2 ^ ")"

    let rec programType =
        function TInt -> TyInt
         | TString -> TyString
         | TBool -> TyBool
         | TPair (t1, t2) -> TyPair (programType t1, programType t2)
         | TLoc t -> TyLoc (programType t)
         | t -> raise (TypeError ("Invalid program type: " ^ string_of_types t))

    (* primitive operators, values *)
    let ( @@ ) g f = (fun t -> g (f t))
    let ( @+ ) g (x, t) = fun y -> if y = x then t else g y
    let emptyG = fun x -> raise (TypeError ("Unknown id: " ^ x))
    let id = (fun x -> x)


    (* substitution, unification part *)
    (* subst.t = ty -> ty *)

    (* subst: int -> ty -> subst *)
    let subst x tau =
        let rec s t = 
			match t with
              TVar y -> if y = x then tau else t
            | TPair (t1, t2) -> TPair (s t1, s t2)
            | TArrow (t1, t2) -> TArrow (s t1, s t2)
            | TLoc t' -> TLoc (s t')
            | TInt | TString | TBool -> t
        in  s

	(* @| : (ty -> ty) -> (id -> ty) -> (id -> ty) *)
    let ( @| ) s g = fun x -> s (g x)

    (* occurs: int -> ty -> bool *)
    let rec occurs x tau = 
		match tau with
          TVar y -> if y = x then true else false
        | TPair (t1, t2) | TArrow (t1, t2) -> occurs x t1 || occurs x t2
        | TLoc t -> occurs x t
        | _ -> false

    (* unify: ty -> ty -> subst *)
    let rec unify t1 t2 =
		match (t1, t2) with
		  (TVar x, tau) -> 
        	if TVar x = tau then id
	        else if not (occurs x tau) then subst x tau
	        else raise (TypeError ("infinite type: " ^
            string_of_types (TVar x) ^ " = " ^ string_of_types tau))
      	| (tau, TVar x) -> unify (TVar x) tau
      	| (TPair (a, b), TPair (c, d)) -> unifypair (a, b) (c, d)
	    | (TArrow (a, b), TArrow (c, d)) -> unifypair (a, b) (c, d)
      	| (TLoc t, TLoc t') -> unify t t'
	    | (tau, tau') ->
	        if tau = tau' then id
		    else raise (TypeError ("mismatch between " ^
	            string_of_types tau ^ " and " ^ string_of_types tau'))
	and unifypair (t1, t2) (t1', t2') = 
        let s = unify t1 t1' in
        let s' = unify (s t2) (s t2') in
        	s' @@ s

    (* constraint derivation and solving part *)

    type constraints = U of exp * ty * ty
                    | Or of exp * ty * ty list

    exception TypeError' of string * exp
    exception AbortCheck of string * exp
	
    (* m: (id -> ty) -> exp -> ty -> (subst * constraints list) *)
    let rec m g exp tau = 
        let mkor tau taul = [Or (exp, tau, taul)] in
		try
        	match exp with
              CONST (S _) -> (unify tau TString, [])
            | CONST (N _) -> (unify tau TInt, [])
            | CONST (B _) -> (unify tau TBool, [])
            | VAR x -> (unify tau (g x), [])
            | FN (x, e) ->
                let t1 = newt() in
                let t2 = newt() in
                let s1 = unify tau (TArrow (t1, t2)) in
                let (s2, c2) = m (s1 @| (g @+ (x, t1))) e (s1 t2) in
                	(s2 @@ s1, c2)
             | APP (e1, e2) ->
                let t1 = newt() in
                let (s1, c1) = m g e1 (TArrow (t1, tau)) in
                let (s2, c2) = m (s1 @| g) e2 (s1 t1) in
                	(s2 @@ s1, c1 @ c2)
             | LET (NREC (x, e1), e2) ->
                let t1 = newt() in
                let g' = g @+ (x, t1) in
                let (s1, c1) = m g' e2 tau in
                let (s2, c2) = m (s1 @| g) e1 (s1 t1) in
                	(s2 @@ s1, c1 @ c2)
             | LET (REC (x, e1), e2) ->
                let t1 = newt() in
                let g' = g @+ (x, t1) in
                let (s1, c1) = m g' e2 tau in
                let (s2, c2) = m (s1 @| g') e1 (s1 t1) in
                	(s2 @@ s1, c1 @ c2)
             | IF (e1, e2, e3) -> 
                let (s1, c1) = m g e3 tau in
                let (s2, c2) = m (s1 @| g) e2 (s1 tau) in
                let s2s1 = s2 @@ s1 in
                let (s3, c3) = m (s2s1 @| g) e1 TBool in
                	(s3 @@ s2s1, c1 @ c2 @ c3)
             | BOP (op, e1, e2) ->
                let (ot, rt) = match op with ADD | SUB -> (TInt, TInt)
                                     | AND | OR -> (TBool, TBool)
                                     | EQ -> (newt(), TBool) in
                let s1 = unify tau rt in
                let (s2, c2) = m g e2 ot in
                let s2s1 = s2 @@ s1 in
                let (s3, c3) = m (s2s1 @| g) e1 (s2s1 ot) in
                let s3s2s1 = s3 @@ s2s1 in
                	(s3s2s1, c2 @ c3 @
                     (mkor (s3s2s1 ot) [TInt; TString; TBool; TLoc (newt())]))
             | READ -> (unify tau TInt, [])
             | WRITE e -> 
                let (s1, c1) = m g e tau in
                	(s1, c1 @ (mkor (s1 tau) [TInt; TString; TBool]))
             | MALLOC e ->
                let t = newt() in
                let s1 = unify tau (TLoc t) in
                let (s2, c2) = m (s1 @| g) e (s1 t) in
                	(s2 @@ s1, c2)
             | ASSIGN (e1, e2) -> 
                let (s1, c1) = m g e2 tau in
                let (s2, c2) = m (s1 @| g) e1 (s1 (TLoc tau)) in
                	(s2 @@ s1, c1 @ c2)
             | BANG e -> m g e (TLoc tau)
             | SEQ (e1, e2) -> 
                let (s1, c1) = m g e2 tau in
                let (s2, c2) = m (s1 @| g) e1 (newt()) in
                	(s2 @@ s1, c1 @ c2)
             | PAIR (e1, e2) ->
                let t1 = newt() in
                let t2 = newt() in
                let s1 = unify tau (TPair (t1, t2)) in
                let (s2, c2) = m (s1 @| g) e2 (s1 t2) in
                let s2s1 = s2 @@ s1 in
                let (s3, c3) = m (s2s1 @| g) e1 (s2s1 t1) in
                	(s3 @@ s2s1, c2 @ c3)
             | SEL1 e -> m g e (TPair (tau, newt()))
             | SEL2 e -> m g e (TPair (newt(), tau))
		with TypeError msg -> raise (TypeError' (msg, exp))

    (* c2s: subst -> constraints list -> subst *)
    let rec c2s s = 
		function [] -> s
      	| (U (e, t1, t2)::c) -> 
			c2s (try (unify (s t1) (s t2) @@ s)
            	with (TypeError msg) -> raise (TypeError' (msg, e))) c
      	| (Or (e, t1, tl)::c) ->
        	let rec tt = 
			function [] -> raise (AbortCheck ("none matches " ^
                    string_of_types (s t1) ^ " among " ^
                    String.concat ", " (List.map (string_of_types @@ s) tl), e))
            | (h::t) -> try c2s s (U (e, t1, h)::c)
                    with TypeError' _ -> tt t
        in  tt tl

    (* check: exp -> types *)
    let check exp = 
        let tau = newt() in
        let (s, c) = m emptyG exp tau in
        	try programType ((c2s s c) tau)
        	with TypeError' (msg, e) | AbortCheck (msg, e) ->
            	raise (TypeError ("For `" ^ M_String.string_of e ^ "', " ^ msg ^ "."))
end

module M_SimChecker : M_SimTypeChecker = struct
    open M

	(* definition of types part *)

    type ty = TVar of int
            | TInt | TString | TBool
            | TPair of ty * ty
            | TLoc of ty
            | TArrow of ty * ty

    let next_tvar = ref 0
    let newt() = next_tvar := !next_tvar + 1; TVar !next_tvar

    let rec string_of_types =
        function TInt -> "int"
         | TString -> "string"
         | TBool -> "bool"
         | TVar x -> "'" ^ string_of_int x
         | TPair (t1, t2) ->
            "(" ^ string_of_types t1 ^ " * " ^ string_of_types t2 ^ ")"
         | TLoc t -> "loc " ^ string_of_types t
         | TArrow (t1, t2) ->
            "(" ^ string_of_types t1 ^ "->" ^ string_of_types t2 ^ ")"

    let rec programType =
        function TInt -> TyInt
         | TString -> TyString
         | TBool -> TyBool
         | TPair (t1, t2) -> TyPair (programType t1, programType t2)
         | TLoc t -> TyLoc (programType t)
         | t -> raise (TypeError ("Invalid program type: " ^ string_of_types t))

    (* primitive operators, values *)
    let ( @@ ) g f = (fun t -> g (f t))
    let ( @+ ) g (x, t) = fun y -> if y = x then t else g y
    let emptyG = fun x -> raise (TypeError ("Unknown id: " ^ x))
    let id = (fun x -> x)


    (* substitution, unification part *)

    (* subst: int -> ty -> subst *)
    let subst x tau =
        let rec s t = 
			match t with
              TVar y -> if y = x then tau else t
            | TPair (t1, t2) -> TPair (s t1, s t2)
            | TArrow (t1, t2) -> TArrow (s t1, s t2)
            | TLoc t' -> TLoc (s t')
            | TInt | TString | TBool -> t
		in s

    (* occurs: int -> ty -> bool *)
    let rec occurs x tau = 
		match tau with
          TVar y -> if y = x then true else false
        | TPair (t1, t2) | TArrow (t1, t2) -> occurs x t1 || occurs x t2
        | TLoc t -> occurs x t
        | _ -> false

	(* unify: ty -> ty -> subst *)
    let rec unify t1 t2 =
		match (t1, t2) with
		  (TVar x, tau) -> 
        	if TVar x = tau then id
	        else if not (occurs x tau) then subst x tau
	        else raise (TypeError ("infinite type: " ^
            string_of_types (TVar x) ^ " = " ^ string_of_types tau))
      	| (tau, TVar x) -> unify (TVar x) tau
      	| (TPair (a, b), TPair (c, d)) -> unifypair (a, b) (c, d)
	    | (TArrow (a, b), TArrow (c, d)) -> unifypair (a, b) (c, d)
      	| (TLoc t, TLoc t') -> unify t t'
	    | (tau, tau') ->
	        if tau = tau' then id
		    else raise (TypeError ("mismatch between " ^
	            string_of_types tau ^ " and " ^ string_of_types tau'))
	and unifypair (t1, t2) (t1', t2') = 
        let s = unify t1 t1' in
        let s' = unify (s t2) (s t2') in
        	s' @@ s


    (* constraint derivation and solving part *)

    type constraints = U of exp * ty * ty
                    | Or of exp * ty * ty list

    exception TypeError' of string * exp
    exception AbortCheck of string * exp

    (* v: (id -> ty) -> exp -> ty -> constraints list -> constraints list *)
    let rec v g exp tau = 
        let u tau tau' = (fun c -> U (exp, tau, tau') :: c) in
        let mkor tau taul = (fun c -> Or (exp, tau, taul) :: c) in
        	match exp with
              CONST (S _) -> u tau TString
            | CONST (N _) -> u tau TInt
            | CONST (B _) -> u tau TBool
            | VAR x -> u tau (g x)
            | FN (x, e) ->
            	let t1 = newt() in
            	let t2 = newt() in
               		u tau (TArrow (t1, t2)) @@ v (g @+ (x, t1)) e t2
            | APP (e1, e2) ->
            	let t1 = newt() in
               		v g e1 (TArrow (t1, tau)) @@ v g e2 t1
            | LET (NREC (x, e1), e2) ->
                let t1 = newt() in
                let g' = g @+ (x, t1) in
                	v g' e2 tau @@ v g e1 t1
            | LET (REC (x, e1), e2) ->
                let t1 = newt() in
                let g' = g @+ (x, t1) in
                	v g' e2 tau @@ v g' e1 t1
             | IF (e1, e2, e3) -> v g e3 tau @@ v g e2 tau @@ v g e1 TBool
             | BOP (op, e1, e2) ->
                let (t, t') = match op with ADD | SUB -> (TInt, TInt)
                                     | AND | OR -> (TBool, TBool)
                                     | EQ -> (newt(), TBool)
                in  
					u tau t' @@ mkor t [TInt; TString; TBool; TLoc (newt())] @@
                    v g e2 t @@ v g e1 t
             | READ -> u tau TInt
             | WRITE e -> mkor tau [TInt; TString; TBool] @@ v g e tau
             | MALLOC e ->
                let t = newt() in
                	u tau (TLoc t) @@ v g e t 
             | ASSIGN (e1, e2) -> v g e2 tau @@ v g e1 (TLoc tau)
             | BANG e -> v g e (TLoc tau)
             | SEQ (e1, e2) -> v g e2 tau @@ v g e1 (newt())
             | PAIR (e1, e2) ->
                let t1 = newt() in
                let t2 = newt() in
                	u tau (TPair (t1, t2)) @@ v g e2 t2 @@ v g e1 t1
             | SEL1 e -> v g e (TPair (tau, newt()))
             | SEL2 e -> v g e (TPair (newt(), tau))

	(* c2s: subst -> constraints list -> subst *)
    let rec c2s s = 
		function [] -> s
      	| (U (e, t1, t2)::c) -> 
			c2s (try (unify (s t1) (s t2) @@ s)
            	with (TypeError msg) -> raise (TypeError' (msg, e))) c
      	| (Or (e, t1, tl)::c) ->
        	let rec tt = 
			function [] -> raise (AbortCheck ("none matches " ^
                    string_of_types (s t1) ^ " among " ^
                    String.concat ", " (List.map (string_of_types @@ s) tl), e))
            | (h::t) -> try c2s s (U (e, t1, h)::c)
                    with TypeError' _ -> tt t
        in  tt tl

    (* reorder: constraints list -> constraints list *)
    let reorder c =
        let (u, o) = List.partition (function Or _ -> false | _ -> true) c in
        	u @ o

    (* check: exp -> types *)
    let check exp = 
        let tau = newt() in
        try programType ((c2s id (reorder (v emptyG exp tau []))) tau)
        with TypeError' (msg, e) | AbortCheck (msg, e) ->
            raise (TypeError ("For `" ^ M_String.string_of e ^ "', " ^ msg ^ "."))
end

module M_LowFat : M_Runner = struct
    open M

    (* domains *)
    type loc = int
    type value = Num of int
               | String of string
               | Bool of bool
               | Loc of loc
               | Pair of value * value
               | Closure of closure
    and closure = fexpr * env
    and fexpr = Fun of id * exp
              | RecFun of id * id * exp
    and env = id -> value
    type memory = int * (loc -> value)

    (* notations (see 5 page in M.pdf) *)
    (* f @+ (x, v)              f[x |-> v]
     * store M (l, v)           M[l |-> v]
     * fetch M l                M(l)
     *)
    let (@+) f (x, v) = (fun y -> if y = x then v else f y)
    let store (l, m) p = (l, m @+ p)
    let fetch (_, m) l = m l
    let malloc (l, m) = (l, (l+1, m))

    (* auxiliary functions *)
    let error msg = raise (RuntimeError msg)
    let op2fun =
        function ADD -> (fun (Num n1,Num n2) -> Num (n1 + n2))
        | SUB -> (fun (Num n1,Num n2) -> Num (n1 - n2))
        | AND -> (fun (Bool b1,Bool b2) -> Bool (b1 && b2))
        | OR ->  (fun (Bool b1,Bool b2) -> Bool (b1 || b2))
        | EQ ->  (fun (v1,v2) -> Bool (v1 = v2))

    let rec printValue =
        function Num n -> print_int n; print_newline()
         | Bool b -> print_endline (if b then "true" else "false")
         | String s -> print_endline s

    let rec eval env mem exp = 
		match exp with
          CONST c ->
            ((match c with S s -> String s | N n -> Num n | B b -> Bool b), mem)
         | VAR x -> (env x, mem)
         | FN (x, e) -> (Closure (Fun (x, e), env), mem)
         | APP (e1, e2) ->
            let (v1, m') = eval env mem e1 in
            let Closure (c, env') = v1 in
            let (v2, m'') = eval env m' e2 in
            	(match c with Fun (x, e) -> eval (env' @+ (x,v2)) m'' e
                        | RecFun (f, x, e) -> eval ((env' @+ (x,v2)) @+ (f,v1)) m'' e)
         | LET (NREC (x, e1), e2) ->
            let (v1, m') = eval env mem e1 in
            	eval (env @+ (x,v1)) m' e2
         | LET (REC (f, e1), e2) ->
            let (v1, m') = eval env mem e1 in
            let Closure (c, env') = v1 in
            	(match c with 
				  Fun (x, e) -> eval (env @+ (f, Closure (RecFun (f, x, e), env'))) m' e2
                | _ -> raise (Invalid_argument "redundant let rec"))
         | IF (e1, e2, e3) ->
            let (Bool b, m') = eval env mem e1 in
            	eval env m' (if b then e2 else e3)
         | BOP (op, e1, e2) ->
            let (v1, m') = eval env mem e1 in
            let (v2, m'') = eval env m' e2 in
            	((op2fun op) (v1,v2), m'')
         | READ ->
            let n = try read_int () with _ -> error "read error" in
            	(Num n, mem)
         | WRITE e ->
            let (v, m') = eval env mem e in
            	printValue v; (v, m')
         | MALLOC e ->
            let (v, m') = eval env mem e in
            let (l, m'') = malloc m' in
            	(Loc l, store m'' (l,v))
         | ASSIGN (e1, e2) ->
            let (v1, m') = eval env mem e1 in
            let Loc l = v1 in
            let (v, m'') = eval env m' e2 in
            	(v, store m'' (l,v))
         | BANG e ->
            let (Loc l, m') = eval env mem e in
            	(fetch m' l, m')
         | SEQ (e1, e2) ->
            let (v1, m1) = eval env mem e1 in
            	eval env m1 e2
         | PAIR (e1, e2) ->
            let (v1, m1) = eval env mem e1 in
            let (v2, m2) = eval env m1 e2 in
            	(Pair (v1, v2), m2)
         | SEL1 e | SEL2 e ->
            let (Pair (v1, v2), m') = eval env mem e in
            	((if exp = SEL1 e then fst else snd) (v1, v2), m')

    let emptyEnv = (fun x -> raise (Invalid_argument ("unknown id: " ^ x)))
    let emptyMem = (0, fun l -> raise (Invalid_argument ("unknown loc: " ^ string_of_int l)))

    let run exp = ignore (eval emptyEnv emptyMem exp)
end
